#pragma CODE 

#include "general.h"
#include "bindec.h"
#include "lcd.h"


// Define symbolic names
#define NO_ERROR 0
#define ERROR    1
#define SEC_MAX  10

// Declare single bits in port P4
sbit P4_2 = P4^2;		//P4.2
sbit P4_3 = P4^3;		//P4.3

// Initialise ASCII-string - located in CODE-segment
char code start_value[] = {"  128"};

// Main function starts here
// The purpose of the program is to demonstrate C51 practised and to gain
// experience in manoeuvring the program development tools.
// The program runs repeatedly, with 1 sec. between each run controlled by a
// software delay.
// Each run an ASCII-string is converted into a binary value. The derived 
// value is send to port P5, which is connected to the LEDs on the 80537-
// board.
// After that, the value is increased and the new value is converted back to 
// an ASCII-string.
// On purpose, a syntax error is introduced in the ASCII-string after each
// few runs, just to see the error handling of the conversion function in 
// work.
// The program also manipulates two bits of port P4. These bits switch on and
// off some of the bulbs on the Laboratory Starter Kit. Thus, the bulbs will 
// start flashing in a systematic way when the program is executed on the 
// 80537-board. 

void main(void) 
{
   //Declare variable in BIT-segment
   bit error_bit = NO_ERROR;	//Bit to indicate error in conversion

   //Declare variables in default segment 
   //- in memory model SMALL this is the DATA-segment
   int rslt;	//Binary result
   int i, sec_count=0;

   //Declare array in XDATA-segment
   char xdata asci[7];	//ASCII-string to convert into binary

   // Initialise LCD
   lcd_init();

   //Move ASCII-string from CODE-segment to XDATA-segment
   for (i=0; i<7; i++) asci[i] = start_value[i];
   
   //Initialise bits in output port P4. On the 80537-board
   //P4.2 and P4.3 open and close two electronic relays.
   P4_2 = 0;	//Turn-on the 230V power supply
   P4_3 = 1;	//Turn-off the two bulbs to the right

   //Initialise all 8 bits in output port P5. On the 80537-
   //board these bits control the LED's
   P5   = 0;	//All LED's switched off


	in t2asc(-32768, asci);

   //Proceed forever
   while (1)
   {
     
     // Convert ASCII-string into binary, and check for errors
     if ( asc2int(asci,&rslt) ) error_bit = NO_ERROR;
     else
     { 
       // Error detected in conversion
       rslt = 1;
       error_bit = ERROR;
     }

     P4_2 = error_bit;		//Error state to bit P4.2
     P4_3 = P4_3^1;		//Toggle bit P4.3
     P5   = rslt;		//Binary result to P5

     // print to display
     lcd_xy(0,0);
     lcd_str(asci);
     
     delay(1000);		//Wait approximately 1 sec.

     rslt *= 10;     		//Increment binary value
     int2asc(rslt,asci);	//Convert into ASCII
     
     rslt = 0;
     
     if (++sec_count==SEC_MAX)
     {
       asci[1]='A';		//Introduce syntax-error in ASCII-string,
       sec_count=0;		//if SEC_MAX seconds elapsed
     }

   }	//end while

}

