#pragma CODE

#include "general.h"
#include "bindec.h"
#include "lcd.h"
#include "dac.h"
#include "keyboard.h"


#define RMS_SAMPLES 20
#define RMS_FORMULA(x)  ((x)/20 )

// current time variables
static xdata int msec=0;
static xdata char sec=0, min=0, hour=0, day=0;

// current status
static xdata struct { long int dac1, dac2; } sum;


// timer interrupt...
void timer2(void) interrupt 5 using 1 {
  static data struct dac_samplepair adc;
  static data unsigned char samples;
  static data struct { long int dac1, dac2; } locsum;
  TF2 = 0;
  P4_0 = 1;
  msec++;
  if (msec>999) {
    msec=0;
    sec++;
    if (sec>59) {
      sec = 0;
      min++;
      if (min>59) {
        min = 0;
        hour++;
        if (hour>23) {
          hour=0;
          day++;
          day&=63;
        }
      }
    }
  }
  dac_input( &adc );
  dac_output( &adc );
  samples++;
  if (samples>RMS_SAMPLES) {
  	 sum = locsum;
    locsum.dac1 = 0;
    locsum.dac2 = 0;
    samples=1;
  }
  locsum.dac1 += ABS(adc.dac1);
  locsum.dac2 += ABS(adc.dac2);
  P4_0 = 0;
}
  

// structure of the control word for timer2 (p.125 in ref)
union t2con {
  struct {
    unsigned t2ps:1;
    unsigned res1:2;
    unsigned t2r:2;
    unsigned t2cm:1;
    unsigned t2i:2;
  } f;
  char c;
};


#define RL_POWER P4_2
#define RL_NEAR P4_3
#define RL_FAR P4_4

void main(void) {
   // 
	char xdata ascii[8];
   char xdata time[] = "  -  .  .  ";
	char count = 0;
   union intchar interval;
   struct dac_samplepair av;
   bit power=0, near=0, far=0;

   // Initialise LCD
   lcd_init();
   lcd_writexy(0, 0,"RMS MULTIMETER");
   lcd_writexy(0, 1,"----------------------------------------");
   lcd_writexy(0, 2,"1KHz samplerate, 20 samples for average");
   lcd_writexy(0, 4,"V(rms) =");
   lcd_writexy(0, 5,"I(rms) =");
   lcd_writexy(0, 8,"Last key =");
   lcd_writexy(0,13,"(A) Power relay:        off");
   lcd_writexy(0,14,"(C) Near short-circuit: off");
   lcd_writexy(0,15,"(F) Far short-circuit:  off");
   

   // Initialise ADC
   dac_init();

	// start keyboard
   kb_init();

   // setup timer2 for sampling
   T2CON |= 0x11;        // timer control word: 0 xx 10 0 01 - no prescaler, reload on overflow, comparemode 0, 
   interval.i = -1000;   // interval for timer2 interrupt
   CRCH = interval.c[0]; //  Reload-vrdi for timer2
   CRCL = interval.c[1]; // -+
	CTCON &= ~0x80;       // T2PS1=0 prescaler level 1 must be zero

   // setup interrupts subsystem   
   IP1 = 0x22;  // set priorities of timer0=2, timer1=3
   IP0 = 0x20;
   IEN1 = 0;    // disable external interrupts
   IEN0 = 0xA2; // allow timer0 and timer2 interrupt, allow all interrupts
   
   // Proceed forever
   while (1) {
     while (msec<500 && !kb_hit()) ;
     time[0] = (day/10)+'0';  time[ 1] = (day%10)+'0';
     time[3] = (hour/10)+'0'; time[ 4] = (hour%10)+'0';
     time[6] = (min/10)+'0';  time[ 7] = (min%10)+'0';
     time[9] = (sec/10)+'0';  time[10] = (sec%10)+'0';
     lcd_writexy(29,0,time);
    
     // print current dac-values
     av.dac1 = RMS_FORMULA(sum.dac1)*10/49;
     av.dac2 = RMS_FORMULA(sum.dac2)*11/237;
     int2asc(av.dac1, ascii);
     lcd_writexy(8,4,ascii);
     int2asc(av.dac2, ascii);
     lcd_writexy(8,5,ascii);
                    
	  while (msec>500 && !kb_hit()) ;
     // print current dac-values
     av.dac1 = RMS_FORMULA(sum.dac1)*10/49;
     av.dac2 = RMS_FORMULA(sum.dac2)*11/237;
     int2asc(av.dac1, ascii);
     lcd_writexy(8,4,ascii);
     int2asc(av.dac2, ascii);
     lcd_writexy(8,5,ascii);

     // empty keyboard-queue
     ascii[1] = 0;
     while (kb_hit()) 
       switch (ascii[0]=kb_readkey()) {
         case 'A' :  // all breakers
           power = !power; 
           RL_POWER = !power; 
           lcd_writexy(24,13,(power?"on ":"off")); 
           break; 
           
         case 'C' : // close breaker
           near = !near; 
           RL_NEAR = !near; 
           lcd_writexy(24,14,(near?"on ":"off")); 
           break; 
           
         case 'F' : // far breaker
           far = !far; 
           RL_FAR = !far; 
           lcd_writexy(24,15,(far?"on ":"off")); 
           break; 
           
         default: 
           lcd_writexy(11,8,ascii);
       }
     
   }	//end while

}

