#pragma ot(2,speed)
#pragma NOMOD517
#include "keyboard.h"

// local keyboard queue
static xdata char queue[8];
static data char start, end;

// enqueue a single keystroke
void kb_enqueue(char c) {
  if ( ((start+1)&0x07) == end) return; // drop enqueue if queue is full
  queue[start] = c; // enqueue the character at the free space
  start = (start+1) & 0x07; // move to next free space in queue
}


// the keyboard-scanner uses register bank 2
/*
 \   (0) (1) (2) (3) 
(a)   1   2   3   F \
(b)   4   5   6   E  |->
(c)   7   8   9   D  |
(d)   A   0   B   C /
      ^

i in pp (the keymap):
  <row0>    <row1>    <row2>    <row3>
  4 1 A 7   5 2 0 8   6 3 B 9   E F C D
  [msb-----------------------------lsb]
*/
union intchar bdata pp;
// correct
sbit key_0 = pp.i^2;
sbit key_1 = pp.i^5;
sbit key_2 = pp.i^1;
sbit key_3 = pp.i^13;
sbit key_4 = pp.i^4;
sbit key_5 = pp.i^0;
sbit key_6 = pp.i^12;
sbit key_7 = pp.i^7;
sbit key_8 = pp.i^3;
sbit key_9 = pp.i^15;
sbit key_A = pp.i^6;
sbit key_B = pp.i^14;
sbit key_C = pp.i^10;
sbit key_D = pp.i^11;
sbit key_E = pp.i^8;
sbit key_F = pp.i^9;


void kb_scan(void) interrupt 1 using 2 {
  static data int oldkeymap;
  static data int newkeymap;
  static char xdata* IO_KEYBOARD = 0xFFFF;
  // stop timer
  TR0 = 0;
  
  // detect any keypress...scan all columns at once
  *(IO_KEYBOARD) = 0xFF;
  pp.c[0] = *(IO_KEYBOARD);
  pp.c[0] &= 0x3C; // 00111100b - pick all relevant bits
  pp.c[0] ^= 0x3C; // 00111100b - if any bit is zero in the four 1-bits, after xor pp !=0
  if ( pp.c[0] || oldkeymap) {
    // reset new keymap
    pp.i = 0;
    *(IO_KEYBOARD) = 0x40; // scan col.0
    pp.c[0] |= (*IO_KEYBOARD &0x3C) << 2;

    *(IO_KEYBOARD) = 0x20; // scan col.1
    pp.c[0] |= (*IO_KEYBOARD &0x3C) >> 2;

    *(IO_KEYBOARD) = 0x02; // scan col.2
    pp.c[1] |= (*IO_KEYBOARD &0x3C) << 2;
    
    *(IO_KEYBOARD) = 0x01; // scan col.3
    pp.c[1] |= (*IO_KEYBOARD &0x3C) >> 2;
    
    newkeymap = ~pp.i;
    pp.i = (newkeymap ^ oldkeymap) & newkeymap;

    if ( key_0 ) kb_enqueue('0');
    if ( key_1 ) kb_enqueue('1');
    if ( key_2 ) kb_enqueue('2');
    if ( key_3 ) kb_enqueue('3');
    if ( key_4 ) kb_enqueue('4');
    if ( key_5 ) kb_enqueue('5');
    if ( key_6 ) kb_enqueue('6');
    if ( key_7 ) kb_enqueue('7');
    if ( key_8 ) kb_enqueue('8');
    if ( key_9 ) kb_enqueue('9');
    if ( key_A ) kb_enqueue('A');
    if ( key_B ) kb_enqueue('B');
    if ( key_C ) kb_enqueue('C');
    if ( key_D ) kb_enqueue('D');
    if ( key_E ) kb_enqueue('E');
    if ( key_F ) kb_enqueue('F');

    oldkeymap = newkeymap;
    // emergency breaker: [E]
    if (newkeymap&0x0001) P4_2=1;
    
  }
  TH0 = 0x3C;
  TL0 = 0xAF;
  TR0 = 1;
}



// Initialise the keyboard driver, start the timer
void kb_init(void) {
  // setup queue to be empty
  start = 0;
  end = 0;
  // set interrupt available
  IP1 |= 0x02;
  IP0 &= ~0x02;
  ET0 = 1;
  EAL = 1;
  // set timer0 mode 01, TR0 bit controls running and as timer
  TMOD = (TMOD & 0xF0) | 0x01;
  // set interval to 50000 (=15535+50000=interrupt)
  TH0 = 0x3C; TL0 = 0xAF;
  TR0 = 1;
}


// Reads single character from keyboard-queue
// Returns value in range '0'..'9','A'..'F' 
// Returns 0 if no keypress is present in keyboard-queue
// Critical section: end = (end+1) & 0x7
//                   this must not be interrupted
//                   so Timer0 is temporarely disabled
char kb_readkey(void) {
//char kb_dequeue(void) {
  char c;
  c = queue[end]; // always read character
  if ( end==start) return 0; // if end = start then queue is empty
  ET0 = 0; 
  end = (end+1) & 0x07; // move to next element in queue
  ET0 = 1;
  return c; // return the previously read character
}

// Is a keypress present in the queue
char kb_hit(void) {
	return end!=start;
}

// Flush all keystrokes, Must be syncronized with Timer0 so it is not interrupted
void kb_flush(void) {
	ET0 = 0;
	start = 0; end = 0;
	ET0 = 1;
}
	

