#pragma CODE 

// 2000-06-20: added lcd_writexy()

#include "lcd.h"
#include "general.h"

#pragma NOMOD517

// Module LCD.C contains functions related to writing of alpha numeric
// text on a dot matrix LCD display.
// The functions are used for initialisation of the display, for positioning
// of the cursor, for writing a string and for converting an integer to
// display code.



// Addresses of registers used to access the display

#define LCD_COMMAND_ADDR        0x0FFE1
#define LCD_STATUS_ADDR         0x0FFE1
#define LCD_WRITE_ADDR          0x0FFE0
#define LCD_READ_ADDR           0x0FFE0


// Prototypes for functions used internally in this module
static void lcd_write_one(char) large;
static void lcd_command(char) large;
static void lcd_write_two(int) large;
static void lcd_auto_write(char) large;


// Pointers to display registers
static char xdata *xdata command_ptr = LCD_COMMAND_ADDR;
static char xdata *xdata status_ptr  = LCD_STATUS_ADDR;
static char xdata *xdata write_ptr   = LCD_WRITE_ADDR;
static char xdata *xdata read_ptr    = LCD_READ_ADDR;


/*********************************************************************/

// Function 'lcd_init' initialises the display to run in alpha numeric
// mode - 16 rows X 40 columns

void lcd_init() large
{
  char xdata i,j;


  lcd_command(0xB2);            //Auto mode reset

  lcd_write_two(0x0000);
  lcd_command(0x40);            //Text Home Addr = 0

  lcd_write_two(0x0028);
  lcd_command(0x41);            //Text Area = 40 columns

  lcd_command(0x80);            //OR Mode

  lcd_write_two(0x0000);
  lcd_command(0x22);            //Offset register set=0

  lcd_write_two(0x0000);
  lcd_command(0x24);            //Address pointer set= 0.0

  lcd_command(0x94);            //Text on, Graphics off, Cursor off

  lcd_command(0xB0);            //Auto mode set

  for (j=0;j<16;j++)
  {
    for (i=0;i<40;i++)          //Blank display
    {
     lcd_auto_write(0x00);
    }
  }

  lcd_command(0xB2);            //Auto mode reset

  lcd_write_two(0x0000);
  lcd_command(0x24);            //Address pointer set= 0.0

}


/*********************************************************************/


// Function 'lcd_command' excecutes a display command by writing
// a byte to the command register.
//      Parameters:
//                  char command;  //display command
//

                       
static void lcd_command(char command) large
{
  while ( (*status_ptr & 0x3) != 0x3); //Wait until ready
  *command_ptr=command;
}

/*********************************************************************/

// Function 'lcd_write_one' prepares for the excecution of a subsequent
// command by writing a byte to the write register.
//      Parameters:
//                  char one;  //Command specification
//

static void lcd_write_one(char one) large
{
  while ( (*status_ptr & 0x3) != 0x3);
  *write_ptr=one;
}

/*********************************************************************/

// Function 'lcd_write_two' prepares for the excecution of a subsequent
// command by writing 2 bytes to the write register.
//      Parameters:
//                  int two;  //Command specification
//

static void lcd_write_two(int two) large
{
  union int_char
  {
    int  var_int;
    char var_char[2];
  };

  union int_char xdata tmp;

  tmp.var_int=two;

  while ( (*status_ptr & 0x3) != 0x3);  //Wait until ready
  *write_ptr=tmp.var_char[1];

  while ( (*status_ptr & 0x3) != 0x3);  //Wait until ready
  *write_ptr=tmp.var_char[0];
}


/*********************************************************************/

// Function 'lcd_auto_write' writes a single characters on the display.
//      Parameters:
//                   char para;  //Character to be written
//

static void lcd_auto_write(char para) large
{
  while ( (*status_ptr & 0x8) != 0x8);  //Wait until ready
  *write_ptr=para;
}



// Since writing on a given position is the most used function,
// lcd_xy() and lcd_str has been merged to form lcd_writexy(x,y,str)
void lcd_writexy(char x, char y, char* str) large {
  char xdata i=0, ch;
  int xdata  pos;

  // ripped from lcd_xy
  lcd_command(0xB2);            // Auto mode reset

  pos = (int)y*40 + x;

  lcd_write_two(pos);
  lcd_command(0x24);            // Address pointer set= pos

  // ripped from lcd_str
  lcd_command(0xB0);            // Auto mode set

  while ( (ch=str[i++]) != '\0' )
    lcd_auto_write(ch-0x20);    //Write characters
  
  lcd_command(0xB2);            //Auto mode reset
}

