; Floating-point library for the SAB 80C537 device using the on-board MDU (Multiply/Divide Unit)
; C51 must use register-calling-convention so pointers are passed in (R1,R2,R3)
;
; Library written by Rune, june 2000
;
; History
; 2000-06-24	history started
; 2000-06-26	the structure is erroneous due to the compiler
;		00000000 SEEEEEEE MMMMMMMM MMMMMMMM
;		internally, only the last 3 bytes are used on the stack
; 2002-07-04	LET ME OUT OF HERE!!!!!!

$NOMOD51
$INCLUDE(REG517.INC)

NAME FP

?PR?fp_mul?FP		SEGMENT CODE
?PR?fp_div?FP		SEGMENT CODE
?PR?fp_add?FP		SEGMENT CODE
?PR?fp_sub?FP		SEGMENT CODE
?PR?_fp_push?FP		SEGMENT CODE
?PR?_fp_pop?FP		SEGMENT CODE
?PR?_fp_pushd?FP	SEGMENT CODE
?PR?_fp_popd?FP		SEGMENT CODE
?PR?fp_popasint?FP	SEGMENT CODE

?DT?fp_stackseg?FP	SEGMENT DATA



; External functions for generic pointer handling - 
EXTRN	CODE (?C?CSTPTR)	; stores A in *(R1,R2,R3)
EXTRN	CODE (?C?CLDPTR)	; loads from *(R1,R2,R3) to A

; All functions return a single result-bit in the carry flag as indication of successful
; computation and must therefore be declared "bit fp_...(void)" 
PUBLIC fp_mul	; declared: bit fp_mul(void)
PUBLIC fp_div	; declared: bit fp_div(void)
PUBLIC fp_add	; declared: bit fp_add(void)
PUBLIC fp_sub	; declared: bit fp_sub(void)
PUBLIC _fp_push	; declared: bit fp_push(struct fp *)
PUBLIC _fp_pushd; declared: bit fp_pushd(struct fp data*)
PUBLIC _fp_pop	; declared: bit fp_pop(struct fp *)
PUBLIC _fp_popd	; declared: bit fp_pop(struct fp data*)
PUBLIC fp_popasint ; declared: int fp_popasint()
PUBLIC fp_stack
PUBLIC fp_stacktop

; declare the stack
; the fp_stack has the format of "struct fp stack[8];" so each element is bit/exp, lo-byte mantissa, hi-byte mantissa
; the fp_stacktop is a bytecount of the stack - so it is a offset to the first free byte in the stack (to avoid multiplying by 3)
RSEG	?DT?fp_stackseg?FP
fp_stack:	ds	8*3
fp_stacktop:	ds	1	; this is an element count, so multiply with 3 before adding to base of fp_stack
				; fp_stacktop points to first free element (overflows not handled)
BIAS		equ	64


;== MUL =======================================================================
; fp_mul - pop and multiply two top elements on stack and push result on stack
;	pop a
;	pop b
;	res.man = a.man * b.man;
;	res.sign = a.sign ^ b.sign;
;	res.exp = a.exp + b.exp - BIAS + 1 - fpnormalize(res.man);
;	push res
; register usage
; 	R0	address of highest operand on stack (a)
;	R1	address of lowest operand on stack (b)
;	R2	address of destination on stack (same as R1)
;	R4	exponent for first
;	R5	exponent for second
;	R6	resulting exponent
	RSEG	?PR?fp_mul?FP
	USING	1
fp_mul:
	; calculate address of lowest (R1=R2) and highest (R0)
	CLR	C
	MOV	B, #3
	MOV	A, fp_stacktop
	DEC	A
	DEC	A
	MUL	AB
	ADD	A, #fp_stack
	MOV	R1, A
	MOV	R2, A
	ADDC	A, #3
	MOV	R0, A
	; load exponents into R4 and R5
	MOV	A, @R0
	MOV	R4, A
	MOV	A, @R1
	MOV	R5, A
	INC	R0
	INC	R1
	; start multiplication in MDU
	MOV	A, @R0	; needs to swap byte-order for this method
	MOV	R6, A
	MOV	A, @R1
	MOV	R7, A
	INC	R0
	INC	R1
	MOV	A, @R0
	MOV	MD0, A
	MOV	A, @R1
	MOV	MD4, A
	MOV	A, R6
	MOV	MD1, A
	MOV	A, R7
	MOV	MD5, A
	; MDU starts multiplying now, wait 4 cycles for retreiving it
	; meanwhile calculate sign and store in R6
	MOV	A, R4
	MOV	B, R5
	XRL	A, B
	ANL	A, #010000000B
	MOV	R6, A
	CLR	C
	; start normalizing in MDU
	MOV	ARCON, #0
	; prepare to store
	MOV	A, R2
	MOV	R1, A
	; start calculating exponent, leaves exponent in Acc
	MOV	A, R4
	SUBB	A, #(BIAS-1)
	ADD	A, B
	ANL	A, #01111111B
	CLR	C
	MOV	B, ARCON
	ANL	B, #00011111B
	SUBB	A, B
	ORL	A, R6
	; store result on stack	
	MOV	@R1, A
	INC	R1
	MOV	@R1, MD3
	INC	R1
	MOV	@R1, MD2
	DEC	fp_stacktop
	CLR	C
	RET


;== DIV =======================================================================
; fp_div - pop two, divide second popped element with first and push result 
;          on stack
;	pop b
;	pop a
;	res.man = a.man / b.man;
;	res.sign = a.sign ^ b.sign;
;	res.exp = a.exp - b.exp + (BIAS + 16 - 1) - fpnormalize(res.man);
;	push res
; register usage
; 	R0	address of highest operand on stack (b)
;	R1	address of lowest operand on stack (a)
;	R2	address of destination on stack (same as R1)
;	R4	exponent for "b.man"
;	R5	exponent for "a.man"
;	R6	resulting exponent
; stack not affected by division by zero!
	RSEG	?PR?fp_div?FP
	USING	1
fp_div:
	; calculate address of lowest (R1=R2) and highest (R0)
	CLR	C
	MOV	B, #3
	MOV	A, fp_stacktop
	DEC	A
	DEC	A
	MUL	AB
	ADD	A, #fp_stack
	MOV	R1, A
	MOV	R2, A
	ADDC	A, #3
	MOV	R0, A
	; load exponents into R4 (a.exp) and R5 (b.exp)
	MOV	A, @R1
	MOV	R4, A
	MOV	A, @R0
	MOV	R5, A
	INC	R0
	INC	R1
	; start division in MDU
	CLR	A
	MOV	MD0, A
	MOV	MD1, A
	MOV	A, @R1	; needs to swap byte-order for this method
	MOV	R6, A
	INC	R1
	MOV	A, @R1
	MOV	MD2, A
	MOV	A, R6
	MOV	MD3, A 
	MOV	A, @R0
	MOV	R7, A
	INC	R0
	MOV	A, @R0
	MOV	MD4, A
	MOV	A, R7
	MOV	MD5, A
	; MDU starts dividing, wait 6 cycles for retreiving it
	; meanwhile calculate sign and store in R6
	MOV	A, R4
	MOV	B, R5
	XRL	A, B
	ANL	A, #010000000B
	MOV	R6, A
	; check for error while dividing
	MOV	A, ARCON
	ANL	A, #01000000B
	JZ	fp_div1
	SETB	C
	RET
fp_div1:	
	; start normalizing in MDU
	MOV	ARCON, #0
	; prepare to store
	MOV	A, R2
	MOV	R1, A
	; start calculating exponent, leaves exponent in Acc
	MOV	A, R4
	SUBB	A, B
	CLR	C
	ADDC	A, #(BIAS+15)
	ANL	A, #01111111B
	CLR	C
	MOV	B, ARCON
	ANL	B, #00011111B
	SUBB	A, B
	ORL	A, R6
	; store result on stack	
	MOV	@R1, A
	INC	R1
	MOV	@R1, MD3
	INC	R1
	MOV	@R1, MD2
	DEC	fp_stacktop
	CLR	C
	RET


;== ADD / SUB =================================================================

; -----------------------------------------------------------------------------
; Helper method - loads two floats from stack to (R2:R3.R4) and (R5:R6.R7)
; param
; return	(R2:R3.R4) - first popped
; 		(R5:R6.R7) - second popped
; 		R1 - address of second popped (usually destination for result) 
fp_load:
	; calculate address of lowest (R1) and highest (R0) operand
	CLR	C
	MOV	B, #3
	MOV	A, fp_stacktop
	DEC	A
	DEC	A
	MUL	AB
	ADD	A, #fp_stack
	MOV	R1, A
	ADDC	A, #3
	MOV	R0, A
	; read the highest operand off the stack
	MOV	A, @R0
	MOV	R2, A
	INC	R0
	MOV	A, @R0
	MOV	R3, A
	INC	R0
	MOV	A, @R0
	MOV	R4, A
	; read the lowest operand off the stack and restore low pointer
	MOV	A, @R1
	MOV	R5, A
	INC	R1
	MOV	A, @R1
	MOV	R6, A
	INC	R1
	MOV	A, @R1
	MOV	R7, A
	DEC	R1
	DEC	R1
	RET


;------------------------------------------------------------------------------
; Helper method - finds the nummerically largest of the two floating-point
; variables stored in (R7;R6.R5) and (R4;R3.R2) (R4 and R7 is exponent)
; params	R2 - exponent (and sign) of first number
;		R3 - MSB of first number
;		R4 - LSB   - -  "  - -
;		R5 - exponent (and sign) of second number
;		R6 - MSB of second number
;		R7 - LSB   - -  "  - -
; returns	R2 - exponent (and sign) of largest number
;		R3 - MSB of largest number
;		R4 - LSB   - -  "  - -
;		R5 - exponent (and sign) of smallest number
;		R6 - MSB of smallest number
;		R7 - LSB   - -  "  - -
; destroys	A, B

fp_swap:
	; check exponents first - if R7 is greater than R4 borrow in subb will occur
	MOV	A, R2
	MOV	B, R5
	ANL	A, #01111111B
	ANL	B, #01111111B
	CLR	C
	SUBB	A, B
	JC	fp_swap_goahead
	JNZ	fp_swap_noswap
	; then check MSB of mantissa
	MOV	A, R3
	MOV	B, R6
	CLR	C
	SUBB	A, B
	JC	fp_swap_goahead
	JNZ	fp_swap_noswap
	; and finally, compare LSB of mantissa
	MOV	A, R4
	MOV	B, R7
	CLR	C
	SUBB	A, B
	JC	fp_swap_goahead
	; (R2:R3.R4) was largest so no swap performed
fp_swap_noswap:
	CLR	C
	RET
fp_swap_goahead:
	; for some reason it turned out that (R2:R3.R4) was smaller (nummerical)
	; than (R5:R6.R7) so we will swap the two variables and return
	MOV	A, R2
	XCH	A, R5
	MOV	R2, A
	MOV	A, R3
	XCH	A, R6
	MOV	R3, A
	MOV	A, R4
	XCH	A, R7
	MOV	R4, A
	SETB	C
	RET


; -----------------------------------------------------------------------------
; fp_add - pop two, add and push result on stack
; 	pop a
;	pop b
;	if (|b|>|a|) 
;		a <=> b
;	res.exp = a.exp
;	
;	
; reg usage	R0 - pointer to highest operand (worthless after reading)
;		R1 - pointer to lowest operand (and destination after calc)
;		R2,R3,R4 - composes first highest operand, after call to 
;		           fp_swap abs. largest operand
;		R5,R6,R7 - composes lowest and abs. smallest operand
	RSEG	?PR?fp_add?FP
	USING	1
fp_add:
	; Load floats into (R2:R3.R4) [first popped] and 
	; (R5:R6.R7) [second popped] and R1 is destination for result
	LCALL	fp_load

fp_add_noload:
	; now, make sure that the largest (nummerically) operand is in 
	; R2:R3.R4 (a swap is indicated by the carry-flag being set)
	LCALL	fp_swap
	
	; is (large.exp - small.exp) > 16 then fp_add_done
	CLR	C
	MOV	A, R2
	SUBB	A, R5
	ANL	A, #01111111B
	CLR	C
	SUBB	A, #16
	JNC	fp_add_done
	CLR	C
	ADDC	A, #16		; shift-right-count
	ORL	A, #00100000B
	MOV	R0, A		; R0 now contains ARCON control word
	; perform the shift in the MDU
	CLR	A
	MOV	MD0, A
	MOV	MD1, A
	MOV	A, R7
	MOV	MD2, A
	MOV	A, R6
	MOV	MD3, A
	MOV	A, R0
	MOV	ARCON, A
	; wait for the MDU to finish shifting (up to 3 cycles, since largest 
	; shift is guaranteed not to be larger than 16!!!
	; while shifting, decide whether subtraction or addition is performed
	MOV	A, R2
	XRL	A, R5
	RLC	A
	JNC	fp_add_add
fp_add_sub:	
	; subtraction is to be performed
	CLR	C
	MOV	A, R4
	; by now the MDU should be finished, so perform the subtraction
	SUBB	A, MD2
	MOV	R4, A
	MOV	A, R3
	SUBB	A, MD3
	MOV	R3, A
	JMP	fp_add_checkzero
fp_add_add: 
	; addition is to be performed
	CLR	C
	MOV	A, R4
	; by now the MDU should be finished, so perform the add
	ADDC	A, MD2
	MOV	R4, A
	MOV	A, R3
	ADDC	A, MD3
	MOV	R3, A
	; if there is carry-out then shift R3.R4 one right and add one to exp.
	JNC	fp_add_checkzero
	RRC	A
	MOV	R3, A
	MOV	A, R4
	RRC	A
	MOV	R4, A
	INC	R2
fp_add_checkzero:
	CJNE	R3, #0, fp_add_normalize
	CJNE	R4, #0, fp_add_normalize
	MOV	A, #40
	MOV	R2, A
	JMP	fp_add_done
fp_add_normalize:
	; we must normalize the number before returning from the function
	CLR	A
	MOV	MD0, A
	MOV	MD1, A
	MOV	A, R4
	MOV	MD2, A
	MOV	A, R3
	MOV	MD3, A
	MOV	ARCON, #0
	; now the MDU is normalizing, so just wait 3 cycles (18 shifts tops)
	NOP
	CLR	C
	MOV	A, R2
	MOV	B, ARCON
	ANL	B, #00011111B
	SUBB	A, B
	; A contains new exponent, so store all
	MOV	@R1, A
	INC	R1
	MOV	A, MD3
	MOV	@R1, A
	INC	R1
	MOV	A, MD2
	MOV	@R1, A
	DEC	fp_stacktop
	CLR	C
	RET	
fp_add_done:
	; R2,R3,R4 contains result, so write back to stack
	MOV	A, R2
	MOV	@R1, A
	INC	R1
	MOV	A, R3
	MOV	@R1, A
	INC	R1
	MOV	A, R4
	MOV	@R1, A
	DEC	fp_stacktop
	CLR	C
	RET




; -----------------------------------------------------------------------------
; fp_sub - pop two, subtract first from second and push result on stack
	RSEG	?PR?fp_sub?FP
	USING	1
fp_sub:
	LCALL	fp_load
	MOV	A, R2
	XRL	A, #10000000B
	MOV	R2, A
	JMP	fp_add_noload

; -----------------------------------------------------------------------------
; _fp_push - push element on stack
; param		R3 - memory type
;		R2 - MSB of address
;		R1 - LSB of address
; 		the generic pointer points to a structure of type fp
; destroys	A, B, C, DPTR, R0, R1, R2, R3, R7
; return	success pushing var
	RSEG	?PR?_fp_push?FP
	USING	1
_fp_push:
	; calculate address of first free byte on stack and store in R0
	MOV	A, fp_stacktop
	MOV	B, #3
	MUL	AB
	ADD	A, #fp_stack
	MOV	R0, A 
	; loose first byte
	LCALL	_fp_push_incptr
	; read the structure into (R5:R6.R7)
	LCALL	?C?CLDPTR	; load exponent, store in R5
	MOV	R5, A
	LCALL	_fp_push_incptr
	LCALL	?C?CLDPTR	; load MSB, store in R6
	MOV	R6, A
	LCALL	_fp_push_incptr
	LCALL	?C?CLDPTR	; load LSB, store in R7
	MOV	R7, A
	; if bit7 of MSB is set, then number is already normalized!
	MOV	A, R6
	ANL	A, #10000000b
	JNZ	_fp_push_store
	; since bit7 of MSB was 0, we need to do a normalize 
	; in the MDU (at most 18 shifts = 3 cycles)
	CLR	A
	MOV	MD0, A
	MOV	MD1, A
	MOV	A, R7
	MOV	MD2, A
	MOV	A, R6
	MOV	MD3, A
	MOV	ARCON, #0
	; MDU is now normalizing, so prep for the result
	CLR	C
	MOV	A, R5
	NOP
	MOV	B, ARCON
	ANL	B, #00011111b
	SUBB	A, B
	; Acc contains resulting exponent, so go ahead and store number
	MOV	@R0, A
	INC	R0
	MOV	A, MD3
	MOV	@R0, A
	INC	R0
	MOV	A, MD2
	MOV	@R0, A
	JMP	_fp_push_done
_fp_push_store: ; stores number in R5:R6.R7 to stack!
	MOV	A, R5
	MOV	@R0, A
	INC	R0
	MOV	A, R6
	MOV	@R0, A
	INC	R0
	MOV	A, R7
	MOV	@R0, A
_fp_push_done: ; all storing is done, so stack updated and return from subr.
	; step one ahead	
	INC	fp_stacktop
	; end of _fp_push
	RET
; small method to increase generic pointer in (R1,R2,R3)
_fp_push_incptr:
	CLR	C		; move pointer ahead
	MOV	A, R1
	ADDC	A, #1
	MOV	R1, A
	MOV	A, R2
	ADDC	A, #0
	MOV	R2, A
	RET


; -----------------------------------------------------------------------------
; _fp_pushd - push element stored in DATA on stack!!!
; param		R7 - address of source struct, moved to R1 on entry!!!
; destroys	A, B, C, DPTR, R0, R1, R2, R3, R7
; return	success pushing var
	RSEG	?PR?_fp_pushd?FP
	USING	1
_fp_pushd:
	; move address of input to R1
	MOV	A, R7
	MOV	R1, A
	; calculate address of first free byte on stack and store in R0
	MOV	A, fp_stacktop
	MOV	B, #3
	MUL	AB
	ADD	A, #fp_stack
	MOV	R0, A 
	; loose first byte
	INC	R1
	; read exponent
	MOV	A, @R1
	MOV	R5, A
	INC	R1
	; read MSB 
	MOV	A, @R1
	MOV	R6, A
	INC	R1
	; read LSB
	MOV	A, @R1
	MOV	R7, A
	; if bit7 of MSB is set, then number is already normalized!
	MOV	A, R6
	ANL	A, #10000000b
	JNZ	_fp_pushd_store
	; since bit7 of MSB was 0, we need to do a normalize 
	; in the MDU (at most 18 shifts = 3 cycles)
	CLR	A
	MOV	MD0, A
	MOV	MD1, A
	MOV	A, R7
	MOV	MD2, A
	MOV	A, R6
	MOV	MD3, A
	MOV	ARCON, #0
	; MDU is now normalizing, so prep for the result
	CLR	C
	MOV	A, R5
	NOP
	MOV	B, ARCON
	ANL	B, #00011111b
	SUBB	A, B
	; Acc contains resulting exponent, so go ahead and store number
	MOV	@R0, A
	INC	R0
	MOV	A, MD3
	MOV	@R0, A
	INC	R0
	MOV	A, MD2
	MOV	@R0, A
	JMP	_fp_pushd_done
_fp_pushd_store: ; stores number in R5:R6.R7 to stack!
	MOV	A, R5
	MOV	@R0, A
	INC	R0
	MOV	A, R6
	MOV	@R0, A
	INC	R0
	MOV	A, R7
	MOV	@R0, A
_fp_pushd_done: ; all storing is done, so stack updated and return from subr.
	; step one ahead	
	INC	fp_stacktop
	; end of _fp_push
	RET


; -----------------------------------------------------------------------------
; _fp_pop - pop element off stack
;	    fills the area pointed to by (R3;R2.R1) with the stacktop and dec.
;           the stack pointer
;
; Destroys A, B, C, DPTR, R0, R1, R2, R3, R7
;
	RSEG	?PR?_fp_pop?FP
	USING	1
_fp_pop:
	MOV	A, fp_stacktop
	; check for empty stack
	CJNE	A, #0, _fp_pop1
	SETB	C
	RET
	; move to last free stack-pos
_fp_pop1:
	DEC	A
	MOV	fp_stacktop, A
	; calculate address of first byte (stacktop) and place in R0
	MOV	B, #3
	MUL	AB
	ADD	A, #fp_stack
	MOV	R0, A
	; store 0 on first byte of pointer
	CLR	A
	LCALL	?C?CSTPTR
	; copy 3 bytes from stack to passed pointer, loose first byte
	MOV	R7, #3
_fp_pop2:
	CLR	C		; move pointer ahead
	MOV	A, R1
	ADDC	A, #1
	MOV	R1, A
	MOV	A, R2
	ADDC	A, #0
	MOV	R2, A
	MOV	A, @R0		; load byte from stack
	LCALL	?C?CSTPTR	; store to pointer
	INC	R0		; move to next byte on stack
	DJNZ	R7, _fp_pop2
	
	; end of _fp_pop
	CLR	C
	RET


; -----------------------------------------------------------------------------
; _fp_popd - pop element off stack to area in DATA
;	    fills the area pointed to by R7 with the stacktop and dec.
;           the stack pointer
;
; Destroys A, B, C, DPTR, R0, R1, R2, R3, R7
;
	RSEG	?PR?_fp_popd?FP
	USING	1
_fp_popd:
	; move R7 to R1
	MOV	A, R7
	MOV	R1, A
	; check for empty stack
	MOV	A, fp_stacktop
	CJNE	A, #0, _fp_popd1
	SETB	C
	RET
	; move to last free stack-pos
_fp_popd1:
	DEC	A
	MOV	fp_stacktop, A
	; calculate address of first byte (stacktop) and place in R0
	MOV	B, #3
	MUL	AB
	ADD	A, #fp_stack
	MOV	R0, A
	; store 0 on first byte of pointer
	CLR	A
	MOV	@R1, A
	INC	R1
	; copy 3 bytes from stack to passed pointer, loose first byte
	MOV	R7, #3
_fp_popd2:
	MOV	A, @R0		; load byte from stack
	MOV	@R1, A		; write to storage
	INC	R0		; pointers ahead
	INC	R1
	DJNZ	R7, _fp_popd2
	
	; end of _fp_pop
	CLR	C
	RET


; -----------------------------------------------------------------------------
; _fp_popasint - pop element off stack, round off to integer and return
; Returns R6 MSB of int
;         R7 LSB of int
;
; Destroys A, B, C, DPTR, R0, R1, R2, R3, R7
;
	RSEG	?PR?fp_popasint?FP
	USING	1
fp_popasint:
	CLR	A
	MOV	R6, A
	MOV	R7, A
	; check for empty stack
	MOV	A, fp_stacktop
	CJNE	A, #0, fp_popasint1
	RET
	; move to last free stack-pos
fp_popasint1:
	DEC	A
	MOV	fp_stacktop, A
	; calculate address of first byte (stacktop) and place in R0
	MOV	B, #3
	MUL	AB
	ADD	A, #fp_stack
	MOV	R0, A
	; load exponent to R2
	MOV	A, @R0
	MOV	R2, A
	INC	R0
	; reset MDU and load it with data
	MOV	A, @R0
	MOV	R3, A
	INC	R0
	MOV	A, @R0
	MOV	MD0, A
	MOV	A, R3
	MOV	MD1, A
	CLR	A
	MOV	MD2, A
	MOV	MD3, A
	; calculate shift-left count, if shift-right then return 0
	CLR	C
	MOV	A, R2
	ANL	A, #01111111b
	SUBB	A, #40H
	JC	fp_popasint2
	INC	A
	MOV	ARCON, A
	; MDU starts shift-left, now wait for max 3 cycles
	NOP
	NOP
	NOP	
	MOV	A, MD3
	MOV	R6, A
	MOV	A, MD2
	MOV	R7, A
	MOV	A, MD1
	RLC	A
	MOV	A, R7
	ADDC	A, #0
	MOV	R7, A
	MOV	A, R6
	ADDC	A, #0
	MOV	R6, A
	; if signed then calculate two's complement
	MOV	A, R2
	ANL	A, #10000000b
	JZ	fp_popasint2
	CLR	C
	MOV	A, R7
	CPL	A
	ADDC	A, #1
	MOV	R7, A
	MOV	A, R6
	CPL	A
	ADDC	A, #0
	MOV	R6, A
fp_popasint2:	
	; end of _fp_popasint
	CLR	C
	RET


END
