#include "fp.h"

// Floating Points arithmetics in pure ANSI C (with // extension) 
#include <stdlib.h>
#include <stdio.h>

struct FP stack[8];
char stacktop=0;


// normalize FP pointed to
int fpnormalize( long int *r ) {
	int c=0;
	while (!(*r&0x80000000) && (c<32) ) {
		c++;
		*r <<= 1;
	}
	return c;
}


// multiply a by b
struct FP fpmul( struct FP a, struct FP b ) {
	struct FP res;
	long int t = a.man * b.man;
	res.sign = a.sign ^ b.sign;
	res.exp = a.exp + b.exp - BIAS + 1 - fpnormalize(&t);
	res.man = t>>16;	
	return res;
}


// divide a by b
struct FP fpdiv( struct FP a, struct FP b ) {
	struct FP res;
	long int t = ((long int)a.man<<16) / b.man;
	res.sign = a.sign ^ b.sign;
	res.exp = a.exp + 16 - b.exp + BIAS - 1 - fpnormalize(&t);
	res.man = t>>16;
	return res;
}


// add a to b
struct FP fpadd( struct FP a, struct FP b ) {
	struct FP aa, bb, res;
	long int i, j;
	if (a.exp>b.exp) { aa = a; bb = b; } else
	if (a.exp<b.exp) { aa = b; bb = a; } else {
		if (a.man>b.man) { aa = a; bb = b; } else
			aa = b; bb = a;
	} // aa now contains largest of the two, bb contains lowest (nummerical)
	if (aa.exp - bb.exp >= 16) return aa;
	i = (long int)aa.man << 16;
	j = (long int)bb.man << (16-(aa.exp-bb.exp));
	if (aa.sign ^ bb.sign)
		i -= j;
	else
		i += j;
	if (i == 0) { aa.man = 0; aa.exp = BIAS; aa.sign = 0; }
	aa.exp -= fpnormalize(&i); aa.man = i>>16;
	return aa;
}


// subtract b from a
struct FP fpsub( struct FP a, struct FP b ) {
	struct FP aa, bb, res;
	long int i, j;
	if (a.exp>b.exp) { aa = a; bb = b; } 
	else if (a.exp<b.exp) { aa = b; bb = a; aa.sign = !aa.sign; } 
		else {
			if (a.man>b.man) { aa = a; bb = b; } 
			else { aa = b; bb = a; aa.sign = !aa.sign; }
		} 
	// aa now contains largest of the two, bb contains lowest (nummerical)
	if (aa.exp - bb.exp >= 16) return aa;
	i = (long int)aa.man << 16;
	j = (long int)bb.man << (16-(aa.exp-bb.exp));
	if (aa.sign ^ bb.sign) 
		i += j;
	else
		i -= j;
	if (i == 0) { aa.man = 0; aa.exp = BIAS; aa.sign = 0; }
	aa.exp -= fpnormalize(&i); aa.man = i>>16;
	return aa;
}



void fp_add(void) {
	struct FP a,b;
	fp_pop(&a);
	fp_pop(&b);
	a = fpadd(a,b);
	fp_push(&a);
}


void fp_sub(void) {
	struct FP a,b;
	fp_pop(&a);
	fp_pop(&b);
	a = fpsub(b,a);
	fp_push(&a);
}


void fp_mul(void) {
	struct FP a,b;
	fp_pop(&a);
	fp_pop(&b);
	a = fpmul(b,a);
	fp_push(&a);
}
	

void fp_div(void) {
	struct FP a,b;
	fp_pop(&a);
	fp_pop(&b);
	a = fpdiv(b,a);
	fp_push(&a);
}

// push to stack
void fp_push(struct FP *f) {
	stack[stacktop++] = *f;
	stacktop &= 7;
}
	

// pop off stack
void fp_pop(struct FP *f) {
	--stacktop; stacktop &= 7;
	*f = stack[stacktop];
}

	
	
		


