#ifndef FP_H
#define FP_H

#define BIAS 64

// general structure to store floating-point number
// structure is:
// 0000.0000 SEEE.EEEE MMMM.MMMM MMMM.MMMM 
//                     \--MSB--/ \--LSB--/
// the zero's are internally ignored!
struct fp {
	unsigned exp: 7;  // exponent, bias is 64
	unsigned sign: 1; // sign for number
	unsigned man: 16; // mantissa
};

// stack declared in assembler module (backdoor to stack)
//extern struct fp fp_stack[8];
//extern unsigned char fp_stacktop;

// pop two elements, add them and push the result on the stack
bit fp_add(void);

// pop two elements, subtract the first from the second and push result on stack
bit fp_sub(void);

// pop two elements, multiply them and push result on stack
bit fp_mul(void);

// pop two elements (pop a, pop b), divide the second element by the first (b / a) and push result on  stack
bit fp_div(void);

// push to stack
bit fp_push(struct fp *);

// push from DATA-segment to stack
bit fp_pushd(struct fp data *);

// pop off stack
bit fp_pop(struct fp *);

// pop off stack to DATA segment
bit fp_popd(struct fp data *);

// pop off stack as integer, Register-calling-convention
int fp_popasint();

#endif